/* =========================================================================
 * FILE       : vdocapture.c
 * -----------:-------------------------------------------------------------
 * PROGRAMMER : Bo-kyun Na (bkna@npac.syr.edu), IntellActive Research Inc.
 *            : 1999.10.4.
 * -----------:-------------------------------------------------------------
 * ********** :     All rights are reserved by Bo-kyun Na.        **********
 * -----------:-------------------------------------------------------------
 * This example reads an image from /dev/video, transmits compressed video 
 * over the local network, and optionally displays the image on Xwindow.
 * -----------:-------------------------------------------------------------
 * Revision   : 
 * ========================================================================= */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

#include "params.h"		/* parameters */
#include "network.h"		/* network related */
#include "transceive.h"		/* parcket format related */

#include "vdo.h"		/* video related */
#include "h261.h"

unsigned char *clp;

unsigned long wpixel[3];
int expand;

int verbose = 0;

int BV_vdo_init_capture ();
int BV_vdo_init_display ();
void BV_vdo_capture ();
void BV_vdo_display ();


/* *************************************************************************
 * Function    : main ()
 * ------------:------------------------------------------------------------
 * Inputs      : int argc, char *argv[]
 * Outputs     : int 
 * Description : void *Thread_TxVideo (void *arg)
 * ************************************************************************* */
int main (int argc, char **argv)
{
  FILE *fp;
  char fname[80];
  int readable = 0, writable = 0;

  int i;

  /* variables for network */
  Parameters *param;
  TransceiveState *tstate;

  /* variables for video */
  char	dev_vdo[64];
  XWA xwa;
  V4L2 v4l2;
  VIMAGE vimage;
  XEvent event;
  unsigned char *capimg;

  /* variables related to codec  */
  int first = 0;
  int CISsize;
  char inbuf[VDO_WIDTH * VDO_HEIGHT * 3/2]; 
  char outbuf[VDO_WIDTH * VDO_HEIGHT * 3/2];
  char tmpbuf[VDO_WIDTH * VDO_HEIGHT * 3/2];
  unsigned char *y, *u, *v;
  unsigned char *y1, *u1, *v1;
  int userno = 0;

  xwa.capturesize[0].width  = 176;   xwa.capturesize[0].height = 144;
  xwa.capturesize[1].width  = 192;   xwa.capturesize[1].height = 144;
  xwa.capturesize[2].width  = 320;   xwa.capturesize[2].height = 240;
  xwa.capturesize[3].width  = 384;   xwa.capturesize[3].height = 288;
  xwa.capturesize[4].width  = 640;   xwa.capturesize[4].height = 480;
  xwa.capturesize[5].width  = 768;   xwa.capturesize[5].height = 576;
  xwa.capturesize[6].width  = 1024;  xwa.capturesize[6].height = 768;

  /* Put in the device node name */
  strcpy(dev_vdo, VDO_DEVICE);

  /* Initialization of video devices */
  param = (Parameters *) init_TxVidParam (argc, argv);

  /* Initialization of transmission parameters */
  if (param->transmit)
    tstate = (TransceiveState *) initialize_transmit(param, VIDEO);

  /* Pointers for the conversion of RGBtoYUV */
  y = inbuf;
  u = &inbuf[VDO_WIDTH * VDO_HEIGHT];
  v = &inbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  y1 = tmpbuf;
  u1 = &tmpbuf[VDO_WIDTH * VDO_HEIGHT];
  v1 = &tmpbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  xwa.xdepth = 16;		/* init. of depth on xwindow */      
  v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565;

  /* Process command line arguments */
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-C")) {
      if (!strcmp(argv[++i], "H261")) {
	param->codec = H261;
	param->cis_type = "H261";
      }
      else if (!strcmp(argv[i], "H263")) {
	param->codec = H263;
	param->cis_type = "H263";
      }
      else if (!strcmp(argv[i], "JPEG")) {
	param->codec = JPEG;
	param->cis_type = "Jpeg";
      }
      else if (!strcmp(argv[i], "MPEG")) {
	param->codec = MPEG;
	param->cis_type = "Mpeg1";
      }
      else if (!strcmp(argv[i], "RAW")) {
	param->codec = RAW;
	param->cis_type = "RAW";
      }
      else {
	printf ("Specifiy your codec mode\n");
	exit (0);
      }
    }
    if(!strcmp(argv[i], "-D")) {
      strcpy (dev_vdo, argv[++i]);
    }
    if (!strcmp(argv[i], "-i")) {
      strcpy (fname, argv[++i]);
      readable = 1;
      fp = fopen(fname, "r");
      if (fp == NULL) printf ("Can't open image file: %s\n", fname);
      else  printf ("Image file name is %s\n", fname);
    }
    if (!strcmp(argv[i], "-o")) {
      strcpy (fname, argv[++i]);
      writable = 1;
      fp = fopen(fname, "w");
      if (fp == NULL) printf ("Can't open image file: %s\n", fname);
      else  printf ("Image file name is %s\n", fname);
    }
    if(!strcmp(argv[i], "-r")) {
      //      rate = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-s")) {
      //      width = atoi(argv[++i]);
      //      height = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-v")) {
      verbose = 1;
    }
    if(!strcmp(argv[i], "-x")) {
      xwa.xdepth = atoi(argv[++i]);
      if (xwa.xdepth == 15)
	xwa.pixelfmt = V4L2_PIX_FMT_RGB555;
      else if (xwa.xdepth == 16)
	xwa.pixelfmt = V4L2_PIX_FMT_RGB565;
      else if (xwa.xdepth == 24)
	xwa.pixelfmt = V4L2_PIX_FMT_BGR24;
      else if (xwa.xdepth == 32)
	xwa.pixelfmt = V4L2_PIX_FMT_BGR32;
      else {
	printf("Unrecognized display depth. Type correct depth\n");
	xwa.pixelfmt = frame_buffer_format (XtDisplay(xwa.viewport),
			      DefaultRootWindow(XtDisplay(xwa.viewport)), &xwa);
      }
    }
  }
  printf ("codec is %s\n", param->cis_type);

  if (verbose)
    printf("Device node:  %s\n", dev_vdo);

  BV_vdo_init_capture (dev_vdo, &vimage, &v4l2);
  capimg = (unsigned *) malloc (v4l2.fmt.fmt.pix.sizeimage);

  BV_vdo_init_display (argc, argv, &vimage, &v4l2, &xwa);

  if (verbose) {
    printf("Capturing %dx%dx%d \"%4.4s\" images\n", 
	   v4l2.fmt.fmt.pix.width, v4l2.fmt.fmt.pix.height, 
	   v4l2.fmt.fmt.pix.depth, (char *)&v4l2.fmt.fmt.pix.pixelformat);
    printf("Images are %d bytes each\n", v4l2.fmt.fmt.pix.sizeimage);
  }

  init_codec ();		/* Initialization of codec of H261 */
  init_encoder ();		/* Initilization of H261 encoder */
  init_decoder ();


  /* clip table */
  if (!(clp = (unsigned char *) malloc(1024))) error("malloc failed\n");
  clp += 384;
  for (i=-384; i<640; i++) clp[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

  /* Loop processing related to a video frame */
  for (;;) {
    //    sem_wait (&VS);
    //    fprintf (stderr, "SendVidThread ");

    if (XtIsRealized(xwa.viewport)) {

      if (readable)
	fread (vimage.data, v4l2.fmt.fmt.pix.sizeimage, 1, fp);
      else
	read(vimage.vid, capimg, v4l2.fmt.fmt.pix.sizeimage);
      //      memcpy (vimage.data, capimg,v4l2.fmt.fmt.pix.sizeimage);
      //      BV_vdo_display (&vimage, &v4l2, &xwa);

      /* Packetization and sending of video data */
      if (param->transmit) {
	if (param->codec == H261) {
	  /* Conversion image pixel format */
	  RGB16toYUV(capimg, y, u, v, VDO_WIDTH, VDO_HEIGHT);

	  CISsize = ((first %= 10) == 0) ? 
	    p64EncodeIntraFrame (inbuf, outbuf):
	    p64EncodeFrame (inbuf, outbuf);
 
	  /* transmit compressed image sequence data */
	  BV_packetize_vdoframe (outbuf, CISsize, 
				 VDO_WIDTH, VDO_HEIGHT, 
				 tstate, param);

	  /* Decode of recvd message and conversion to RGB from YUV  */
	  decode_frame (outbuf, CISsize, tmpbuf, (first == 0) ? 1 : 0, userno);
	  //	  if (xwa.xdepth == 16)
	  //	    Color16DitherImage(tmpbuf, vimage.data);
	  //	  else if (xwa.xdepth == 32)
	  //	    Color32DitherImage(tmpbuf, vimage.data);

	  ConvertYUVtoRGB(y1, u1, v1, vimage.data, VDO_WIDTH, VDO_HEIGHT);
	  //	memcpy (vimage.data, inbuf, VDO_WIDTH * VDO_HEIGHT * 3/2 );

	  /* Display compressed local images */
	  BV_vdo_display (&vimage, &v4l2, &xwa);
	  first ++;  
	}
	else if (param->codec == H263) {
	  printf ("Not available.\n");
	  BV_vdo_display (&vimage, &v4l2, &xwa);
	}
	else if (param->codec == JPEG) {
	  printf ("Not available.\n");
	  BV_vdo_display (&vimage, &v4l2, &xwa);
	}
	else if (param->codec == MPEG) {
	  printf ("Not available.\n");
	  BV_vdo_display (&vimage, &v4l2, &xwa);
	}
	else if (param->codec == RAW) {	/* transmit raw video data */
	  BV_packetize_vdoframe (capimg, v4l2.fmt.fmt.pix.sizeimage, 
				 VDO_WIDTH, VDO_HEIGHT, tstate, param);

	  /* display local video image */
	  memcpy (vimage.data, capimg, v4l2.fmt.fmt.pix.sizeimage);
	  BV_vdo_display (&vimage, &v4l2, &xwa);
	}
	else {
	  printf ("Error on sending. Specifiy a correct codec type.\n");
	  exit (0);
	}
      }

      /* Store compressed video data as a file */
      if (writable)	
	fwrite(vimage.data, 1, CISsize, fp);
    }
    else {
      XtAppPeekEvent(xwa.app, &event);
    }

    while (XtAppPending(xwa.app)) {
      XtAppNextEvent(xwa.app, &event);
      switch (event.type) {
      default:
	break;
      }
      XtDispatchEvent(&event);
    }

    //    thrd_stop (&AR);   /* Thread stop */
  }

  return 0;
}
