
/* *************************************************************************
 *  File : MTvodo.c
 * -------------------------------------------------------------------------
 *  All rights are reserved by Bo-kyun Na, IntellActive Research Inc.
 * -------------------------------------------------------------------------
 * This example reads an image from /dev/video, transmits compressed video 
 * over the local network, and optionally displays the image in Xwindow.
 * ************************************************************************* */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

#include "params.h"
#include "network.h"
#include "transceive.h"
#include "ext-pthread.h"
#include "vdo.h"		/* video related header */
#include "h261.h"
#include "odo.h"		/* audio related headers */

/* Thread related variables */
pthread_mutex_t utex_lock = PTHREAD_MUTEX_INITIALIZER;
int count = 0;			/* used in thrd_stop () */
sem_t VS, VR, AS, AR;


unsigned char *clp;

unsigned long wpixel[3];
int expand;

int verbose = 0;

void *Thread_TxVideo (void *);
void *Thread_RxVideo (void *);
void *Thread_TxAudio (void *);
void *Thread_RxAudio (void *);



/* *************************************************************************
 * thrd_stop (sem_t *sem)
 *
 * ************************************************************************* */
void thrd_stop (sem_t *sem)
{
  pthread_mutex_lock (&utex_lock);
  (count) ? sem_post (sem), count = 0 : count++;
  pthread_mutex_unlock (&utex_lock);
}


/* *************************************************************************
 * usage ()
 *
 *
 * ************************************************************************* */
void usage (char *pgm)
{
  fprintf (stderr, "usage: %s [options] \n \
Legal options:                                       Defaults\n \
<-i file_name> : audio input file name               0\n \
<-o file_name> : audio output file name              0\n \
", pgm);
}


/* *************************************************************************
 * void *OLD_Thread_TxVideo (void *arg)
 *
 *
 * ************************************************************************* */
void *OLD_Thread_TxVideo (void *arg)
{
  FILE *df;

  //  int frames;
  //  int first_frame = 1;
  unsigned int width, height;
 
  Parameters *param = (Parameters *) arg;
  TransceiveState *tstate;

  int vid;
  int xdepth = 0;
  V4L2 v4l2;
  GC gc_send_img;
  Widget viewport;
  XtAppContext	app;
  char	my_device[64];
  VIMAGE vimage[STREAMBUFS];
  int CISsize;
  char inbuf[VDO_WIDTH * VDO_HEIGHT * 3/2]; 
  char outbuf[VDO_WIDTH * VDO_HEIGHT * 3/2];
  unsigned char *y, *u, *v;


  /* Pointers for the conversion of RGBtoYUV */
  y = inbuf;
  u = &inbuf[VDO_WIDTH * VDO_HEIGHT];
  v = &inbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  /* Put in the device node name */
  strcpy(my_device, VDO_DEVICE);

  /* Initialization of transmission parameters */
  if (param->transmit)
    tstate = (TransceiveState *) initialize_transmit(param, VIDEO);

  /* Function: Init. of video capture to get the video from my camera */
  vid = BV_init_vidcap (my_device, &v4l2);

  v4l2.req.count = 1;		/* Num of requested window displays */

  gc_send_img =			/* Init. and format of xwindow display */
    BV_init_display (&app, &xdepth, &viewport, &v4l2, param->argc, param->argv, vimage);

  init_codec ();		/* Initialization of codec of H261 */
  init_encoder ();		/* Initilization of H261 encoder */

  if (param->datafile)
    df = fopen(param->datafile, "w");

  for (;;) {
    //    sem_wait (&VS);
    fprintf (stderr, "SendVidThread ");

    BV_get_vdoframe (vid, vimage[0].data, &v4l2);
    BV_display_vdoframe (&viewport, &vimage[0], &gc_send_img, &v4l2, &app);
    RGB16toYUV(vimage[0].data, y, u, v, VDO_WIDTH, VDO_HEIGHT);

    /* Implemenation of H261 encoder */
    CISsize = p64EncodeIntraFrame (inbuf, outbuf);
    /*	CISsize = p64EncodeFrame (inbuf, outbuf); */

    if (param->datafile)	/* Store compressed video data as a file */
      fwrite(vimage[0].data, 1, CISsize, df);

    if (param->transmit)     /* Packetization and sending of video data */
      BV_packetize_vdoframe (outbuf, CISsize, VDO_WIDTH, VDO_HEIGHT, tstate, param);

    //      BV_packetize_vdoframe (vimage[0].ximage->data, sizeof (vimage[0].ximage->data), width, height, tstate, param);    
    //    thrd_stop (&AR); 
    /*   thrd_stop (&AS); */
  }
}



/* *************************************************************************
 * Function    : void *Thread_TxVideo (void *arg)
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
//int main (int argc, char **argv)
void *Thread_TxVideo (void *arg)
{
  FILE *fp;
  char fname[80];
  int readable = 0, writable = 0;

  int i;

  Parameters *param;
  TransceiveState *tstate;

  int vid;
  char	devname[64];
  unsigned char *capimg;	/* buffer for captured image */
  XWA xwa;
  V4L2 v4l2;
  VIMAGE vimage;		/* func. can't pass & return struct,    */
				/* instead,  fun. can do pt. of struct  */	

  int first = 0;
  int CISsize;
  char inbuf[VDO_WIDTH * VDO_HEIGHT * 3/2]; 
  char outbuf[VDO_WIDTH * VDO_HEIGHT * 3/2];
  char tmpbuf[VDO_WIDTH * VDO_HEIGHT * 3/2];
  unsigned char *y, *u, *v;
  unsigned char *y1, *u1, *v1;
  int userno = 0;


  /* Put in the device node name */
  strcpy(devname, VDO_DEVICE);

  /* Initialization of video devices */
  param = (Parameters *) init_TxVidParam (argc, argv);

  /* Initialization of transmission parameters */
  if (param->transmit)
    tstate = (TransceiveState *) initialize_transmit(param, VIDEO);

  /* Pointers for the conversion of RGBtoYUV */
  y = inbuf;
  u = &inbuf[VDO_WIDTH * VDO_HEIGHT];
  v = &inbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  y1 = tmpbuf;
  u1 = &tmpbuf[VDO_WIDTH * VDO_HEIGHT];
  v1 = &tmpbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  xwa.xdepth = 16;		/* init. of depth on xwindow */      

  printf ("codec is %s\n", param->cis_type);

  /* Function: Init. of video capture to get the video from my camera */
  vid = BV_init_vidcap (devname, &v4l2);

  /* Init. and format of xwindow display */
  vimage.data = BV_init_display (&xwa, &v4l2, argc, argv, &vimage);
  if (vimage.data == NULL) {
    printf("malloc(%d) for vimage.data failed\n", v4l2.fmt.fmt.pix.sizeimage);
    exit (1);
  }

  capimg = (unsigned char *) malloc (v4l2.fmt.fmt.pix.sizeimage);
  if (capimg == NULL) {
    printf("malloc(%d) for capimg failed\n", v4l2.fmt.fmt.pix.sizeimage);
    exit (1);
  }

  /* Initialization of codec functions */
  init_codec ();		/* Initialization of codec of H261 */
  init_encoder ();		/* Initilization of H261 encoder */
  init_decoder ();

  /* clip table */
  if (!(clp = (unsigned char *) malloc(1024))) error("malloc failed\n");
  clp += 384;
  for (i=-384; i<640; i++) clp[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

  /* Loop processing related to a video frame */
  for (;;) {
    //    sem_wait (&VS);
    //    fprintf (stderr, "SendVidThread ");

    if (readable)
      fread (vimage.data, v4l2.fmt.fmt.pix.sizeimage, 1, fp);
    else
      BV_get_vdoframe (vid, capimg, &v4l2);

    /* Packetization and sending of video data */
    if (param->transmit) {
      if (param->codec == H261) {
	/* Conversion image pixel format */
	RGB16toYUV(capimg, y, u, v, VDO_WIDTH, VDO_HEIGHT);

	CISsize = ((first %= 10) == 0) ? 
	          p64EncodeIntraFrame (inbuf, outbuf):
	          p64EncodeFrame (inbuf, outbuf);
 
	/* transmit compressed image sequence data */
	BV_packetize_vdoframe(outbuf,CISsize,VDO_WIDTH,VDO_HEIGHT,tstate,param);

	/* Decode of recvd message and conversion to RGB from YUV  */
	decode_frame (outbuf, CISsize, tmpbuf, (first == 0) ? 1 : 0, userno);

	ConvertYUVtoRGB(y1, u1, v1, vimage.data, VDO_WIDTH, VDO_HEIGHT);

	/* Display compressed local images */
	BV_display_vdoframe (&xwa, &vimage, &v4l2);

	first ++;  
      }
      else if (param->codec == H263) {
	printf ("Not available.\n");
	exit (0);
      }
      else if (param->codec == JPEG) {
	printf ("Not available.\n");
	exit (0);
      }
      else if (param->codec == MPEG) {
	printf ("Not available.\n");
	exit (0);
      }
      else if (param->codec == RAW) {	/* transmit raw video data */
	BV_packetize_vdoframe (capimg, v4l2.fmt.fmt.pix.sizeimage, 
			       VDO_WIDTH, VDO_HEIGHT, tstate, param);

	/* display local video image */
	memcpy (vimage.data, capimg, v4l2.fmt.fmt.pix.sizeimage);
	BV_display_vdoframe (&xwa, &vimage, &v4l2);
      }
      else {
	printf ("Error on sending. Specifiy a correct codec type.\n");
	exit (0);
      }
    }

    /* Store compressed video data as a file */
    if (writable)	
      fwrite(vimage.data, 1, CISsize, fp);

    //    thrd_stop (&AR);   /* Thread stop */
  }
}


/* ********************************************************************
 * Thread : void *OLD_Thread_RxVideo(void *arg)
 *
 *
 * ******************************************************************** */
void *OLD_Thread_RxVideo (void *arg)
{
  //  int vid;
  int recvflag;
  int xdepth = 0;
  int CISsize, first=1, userno = 0;
  uchar *y, *u, *v;
  char inbuf[VDO_WIDTH * VDO_HEIGHT * 3/2]; 
  char outbuf[VDO_WIDTH * VDO_HEIGHT * 3/2];
  V4L2 v4l2;
  VIMAGE vimage[STREAMBUFS];
  Widget viewport;
  XtAppContext	app;
  GC gc_recvd_img;

  int packetcnt = 0;
  int currentframe;
  int lastw, lasth;
  int i;
  unsigned int currentsession;
  unsigned short myflags;

  Parameters *params = (Parameters *) arg;
  TransceiveState *receiver;
  PacketType packets[MAXPACKETS];	/* circular buffer of rec'vd packets */
  PacketType *packetlist[MAXPACKETS];	/* packets for this frame */
  PacketType *curpacket;
  MsgType message;


  /* .... initialize everything ....  */
  receiver = initialize_receive(params, VIDEO);

  y = inbuf;
  u = &inbuf[VDO_WIDTH * VDO_HEIGHT];
  v = &inbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  receiver->session = 1;	/* choose remote host to receive */
  v4l2.req.count = 1;		/* Num of requested window displays */

  /* 
  (char *) receiver->name[receiver->session].sin_addr = 
                                  (char *) params->hostname[receiver->session];
  printf ("remotehost = %s\n", receiver->name[receiver->session]);
  */

  /* Initialization of image display  */
  gc_recvd_img = BV_init_display (&app, &xdepth, &viewport, &v4l2, 
				  params->argc, params->argv, vimage);

  /* Initilization of H261 decoder */
  init_codec ();		/* Initialization of codec of H261 */
  init_decoder ();	

  /* .... receive video ....      */
  currentframe = 0;
  currentsession = 0;
  lastw = params->width;
  lasth = params->height;
   
  myflags = 0xffff;
  for (i = 0; i < MAXPACKETS; i++) packetlist[i] = NULL;

  for (;;) {
    //    sem_wait (&VR);
    //    sem_post (&VS); 

    fprintf (stderr, "RecvVidThread ");

    curpacket = &packets[packetcnt++];
    if (packetcnt >= MAXPACKETS) packetcnt = 0;
    printf ("Rx:1\n");
    /* Get video packet from remote if available. */
    recvflag = receive_packet(receiver, curpacket);
    printf ("Rx:0\n");
    if(!recvflag) {
      thrd_stop (&AR); 
      continue;
    }
printf ("Rx:2\n");

    if (currentsession != curpacket->header.session) {
      /* if you join a conference in progress then the current frame and 
       * session will be grabbed from the first received packet received.  */
      if (verbose) {
	fprintf(stderr, "changing session from %d to %d\n",
		currentsession, curpacket->header.session);
      }
      for (i = 0; i < MAXPACKETS; i++) 	packetlist[i] = NULL;
      myflags = 0xffff;
      currentframe = curpacket->header.frame;
      currentsession = curpacket->header.session;
    }

    /* Resize of window :
     * if this is the first packet of a frame, check for a resize.
     * In BuenaVista, it should be removed. (x and y in packe header).
     */
    if ((curpacket->header.flags & 1) == 0) {
      /* the size of each frame is sent on the first packet of each frame */
      if ((curpacket->header.x != lastw) || (curpacket->header.y != lasth)) {
	lastw = curpacket->header.x;
	lasth = curpacket->header.y;

	if (verbose) 
	  printf("resizing video %dx%d\n",lastw, lasth);

	/* Resize of window */
	/*	resize_window (window, lastw, lasth); */
      }
    }

    if (curpacket->header.frame < currentframe) {
      /* if this packet is from an earlier frame, disregard it. */
      if (verbose)
	fprintf(stderr, "received stale packet %d, current frame %d\n",
		curpacket->header.frame, currentframe);
      continue;
    }

    if (curpacket->header.frame > currentframe) {
      /* If the frame number increased then we probably missed the last
       * packet of the previous frame. We display what we can drop of the
       * previous frame if '-b' option for verbose was selected, otherwise 
       * we just drop the partial frame.                                  
       */
      if (myflags != 0) {
	if (verbose)
	  fprintf(stderr, "dropping %d due to getting packet %d (flags %x)\n",
		  currentframe, curpacket->header.frame, myflags);
      }

      for (i = 0; i < MAXPACKETS; i++) packetlist[i] = NULL;
      myflags = 0xffff;
      currentframe = curpacket->header.frame;
    }

    /* Save the packet for decoding */
    packetlist[packetnum(curpacket->header.flags)] = curpacket;
    myflags &= curpacket->header.flags;

    if (myflags == 0) {
      if (params->do_cis == H261) {
      	create_message (&message, packetlist);
      }
      else 
      	create_message(&message, packetlist);

    //    decode_frame (inbuf, CISsize, curpacket->payload, first, userno);
    //    YUVtoRGB(vimage[1].data, y, u, v, VDO_WIDTH, VDO_HEIGHT);

      //      BV_display_vdoframe (&viewport, &vimage[1], &gc_recvd_img, &v4l2, &app);
      //      BV_display_vdoframe (&viewport, inbuf, &gc_recvd_img, &v4l2, &app);

      /* Display raw image data */
      vimage[0].ximage->data = curpacket->payload; 
      BV_display_vdoframe (&viewport, &vimage[0], &gc_recvd_img, &v4l2, &app);
    }

    //    thrd_stop (&AR); 
    /*    thrd_stop (&AS); */
  }
}


/* *************************************************************************
 * Function    : void *Thread_RxVideo (void *arg)
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
// void main (int argc, char **argv)
void *Thread_RxVideo (void *arg)
{
  FILE *df;
  char datafile[80];

  int userno = 0;
  uchar *y, *u, *v;
  char inbuf[VDO_WIDTH * VDO_HEIGHT * 3/2]; 

  V4L2 v4l2;
  XWA xwa;
  VIMAGE vimage;

  int packetcnt = 0;
  int currentframe;
  int lastw, lasth;
  int i;
  unsigned int currentsession;
  unsigned short myflags;

  Parameters *param;
  TransceiveState *receiver;
  PacketType packets[MAXPACKETS];	/* circular buffer of rec'vd packets */
  PacketType *packetlist[MAXPACKETS];	/* packets for this frame */
  PacketType *curpacket;
  MsgType message;

  xwa.xdepth = 16;
  v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565;

  /* .... initialize everything ....  */
  param = (Parameters *) init_RxVidParam(argc, argv);
  receiver = initialize_receive(param, VIDEO);

  y = inbuf;
  u = &inbuf[VDO_WIDTH * VDO_HEIGHT];
  v = &inbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  param->data_via_file = 0;	/* Save data to a file */
  receiver->session = 1;	/* choose remote host to receive */
  param->codec = RAW;

  /* 
  (char *) receiver->name[receiver->session].sin_addr = 
                                  (char *) param->hostname[receiver->session];
  printf ("remotehost = %s\n", receiver->name[receiver->session]);
  */

  /* Process command line arguments */
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-C")) {
      if (!strcmp(argv[++i], "H261")) {
      	param->codec = H261;
      	param->cis_type = "H261";
      }
      else if (!strcmp(argv[i], "H263")) {
      	param->codec = H263;
      	param->cis_type = "H263";
      }
      else if (!strcmp(argv[i], "JPEG")) {
      	param->codec = JPEG;
      	param->cis_type = "Jpeg";
      }
      else if (!strcmp(argv[i], "MPEG")) {
      	param->codec = MPEG;
      	param->cis_type = "Mpeg1";
      }
      else if (!strcmp(argv[i], "RAW")) {
	param->codec = RAW;
      	param->cis_type = "RAW";
      }
      else {
	printf ("Specifiy your codec mode\n");
	exit (0);
      }
    }
    /*    if (!strcmp(argv[i], "-i")) {
      strcpy (fname, argv[++i]);
      readable = 1;
      fp = fopen(fname, "r");
      if (fp == NULL) printf ("Can't open image file: %s\n", fname);
      else  printf ("Image file name is %s\n", fname);
    }
    if (!strcmp(argv[i], "-o")) {
      strcpy (fname, argv[++i]);
      writable = 1;
      fp = fopen(fname, "w");
      if (fp == NULL) printf ("Can't open image file: %s\n", fname);
      else  printf ("Image file name is %s\n", fname);
    }
    */
    if(!strcmp(argv[i], "-r")) {
      //      rate = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-s")) {
      //      width = atoi(argv[++i]);
      //      height = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-v")) {
      verbose = 1;
    }
    if(!strcmp(argv[i], "-x")) {
      xwa.xdepth = atoi(argv[++i]);
      if (xwa.xdepth == 15)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB555;
      else if (xwa.xdepth == 16)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565;
      else if (xwa.xdepth == 24)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24;
      else if (xwa.xdepth == 32)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR32;
      else {
	printf("Unrecognized display depth. Type correct depth\n");
	exit (0);
      }
    }
  }

  printf ("codec is %s\n", param->cis_type);

  /* Initialization of image display  */
  vimage.data = BV_init_display (&xwa, &v4l2, argc, argv, &vimage);

  init_codec ();		/* Initialization of codec of H261 */
  init_decoder ();		/* Initilization of H261 decoder */

  /* clip table */
  if (!(clp = (unsigned char *) malloc(1024))) error("malloc failed\n");
  clp += 384;
  for (i=-384; i<640; i++) clp[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

  if (param->data_via_file) {
    strcpy (datafile, "vimage.raw");
    printf ("image file name is %s\n", datafile);
    df = fopen(datafile, "w");
  }

  /* .... receive video ....      */
  currentframe = 0;
  currentsession = 0;
  receiver->session = 0;
  lastw = param->width;
  lasth = param->height;
   
  myflags = 0xffff;
  for (i = 0; i < MAXPACKETS; i++) packetlist[i] = NULL;

  for (;;) {
    //    sem_wait (&VR);
    //    sem_post (&VS); 

    //   fprintf (stderr, "RecvVidThread ");

    curpacket = &packets[packetcnt++];
    packetcnt %= MAXPACKETS; 

    /* Get video packet from remote if available. */
#ifdef XEVENT
    receive_packet(receiver, curpacket, &viewport);
#else
     receive_packet(receiver, curpacket);
#endif

     if (verbose)
       printf ("cursession = %d\n", curpacket->header.session);

    if (curpacket->header.frame < currentframe) {
      /* if this packet is from an earlier frame, disregard it. */
      if (verbose)
	fprintf(stderr, "received stale packet %d, current frame %d\n",
		curpacket->header.frame, currentframe);
      //    thrd_stop (&AR); 
      continue;
    }

    if (curpacket->header.frame > currentframe) {
      if (myflags != 0) {
	if (verbose)
	  fprintf (stderr, "dropping %d due to getting packet %d (flags %d)\n",
		   currentframe, curpacket->header.frame, myflags);
      }

      for (i = 0; i < MAXPACKETS; i++) 
	packetlist[i] = NULL;
      myflags = 0xffff;
      currentframe = curpacket->header.frame;
    }

    /* Save the packet for decoding */
    packetlist[packetnum(curpacket->header.flags)] = curpacket;
    myflags &= curpacket->header.flags;

    if (myflags == 0) {
      if (param->codec == H261) {
	create_message (&message, packetlist);

      /* Decode of recvd message and conversion to RGB from YUV  */
	decode_frame (message.data, message.header.size, inbuf, 
		      ((curpacket->header.frame % 10) == 0) ? 1 : 0, userno);
	ConvertYUVtoRGB(y, u, v, vimage.data, VDO_WIDTH, VDO_HEIGHT);
	//	memcpy(vimage.data, message.data, message.header.size); 
      } 
      else if (param->codec == H263) {
	create_message(&message, packetlist);
      }
      else if (param->codec == JPEG) {
	create_message(&message, packetlist);
      }
      else if (param->codec == MPEG) {
	create_message(&message, packetlist);
      }
      else if (param->codec == RAW) {
	create_message(&message, packetlist);
	memcpy(vimage.data, message.data, message.header.size); 
      }
      else {
	printf ("Error on creating message. Specify the codec type\n");
	exit (0);
      }

      //      if (verbose)
	printf ("Rx: The %dth image frame w/ size = %d bytes\n", 
		message.header.frame, message.header.size);

      /* Display images */
      BV_display_vdoframe (&xwa, &vimage, &v4l2);

      /* Store received video data as a file */
      if (param->data_via_file)	
	fwrite(vimage.data, 1, message.header.size, df);

    }
    //    thrd_stop (&AR); 
  }
}


/* *************************************************************************
 * Function    : Thread : void *Thread_TxAudio(void *arg)
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void *OLD_Thread_TxAudio (void *arg)
{
  TransceiveState *TxState;
  Parameters *param;
  
  param = (Parameters *) arg;

  TxState = (TransceiveState *) initialize_transmit (param, AUDIO);

  for ( ; ; ) {
    //    sem_wait (&AS);

    fprintf (stderr, "SendAudThread\n");



    //    thrd_stop (&VR);  
  }


}

// int main(int argc, char **argv)
void *Thread_TxAudio (void *arg)
{
  int fd;			/* sound device file descriptor */
  int status;			/* return status of system calls */
  int rd;			/* size in byte unit */
  int odo_frame;

  char dev_odo[64];		/* device driver of audio device */
  unsigned char *buf;		/* this buffer holds the digitized audio */

  /* Network parameters */
  Parameters *param;
  TransceiveState *tstate;


  /* audio device driver */
  strcpy (dev_odo, ODO_DEVICE);		

  odo_frame = ODO_LENGTH * ODO_RATE * ODO_SIZE * ODO_CHANNELS / 8;
  buf = (unsigned char *) malloc(odo_frame);
  if (buf == NULL) {
    printf ("malloc fail: sizeof (buf) = %d\n", odo_frame);
    exit (1);
  }

  /* Initialization of network parameters */
  param = (Parameters *) init_TxVidParam (argc, argv);

  /* Initialization of transmission parameters */
  if (param->transmit)
    tstate = (TransceiveState *) initialize_transmit(param, AUDIO);

  /* Initialization of audio */
  fd = BV_ODO_initialize (dev_odo);

  for ( ; ; ) {			/* loop until Control-C */
    rd = BV_ODO_record (fd, buf, odo_frame);
    printf ("Recorded audio size is %d\n", rd);

    BV_odo_packetize (buf, odo_frame, tstate, param);

    //    BV_ODO_play (fd, buf, odo_frame);

    /* wait for playback to complete before recording again */
    status = ioctl(fd, SOUND_PCM_SYNC, 0); 
    if (status == -1)
      perror("SOUND_PCM_SYNC ioctl failed");
  }
}



/* *************************************************************************
 * Function    : Thread : void *Thread_RxAudio(void *arg)
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void *OLD_Thread_RxAudio (void *arg)
{

  Parameters *param = (Parameters *) arg;
  TransceiveState *RxState;

  RxState = (TransceiveState *) initialize_receive (param, AUDIO);

  for ( ; ; ) {
    //    sem_wait (&AR);
    //    sem_post (&AS);

    fprintf (stderr, "RecvAudThread ");



    //    thrd_stop (&VR);
  }

}


// int main(int argc, char **argv)
void *Thread_RxAudio (void *arg)
{
  int fd;			/* sound device file descriptor */
  int status;			/* return status of system calls */
  int wr;			/* size in byte unit */
  int odo_frame;

  char dev_odo[64];		/* device driver of audio device */
  unsigned char *buf;		/* this buffer holds the digitized audio */

  Parameters *param;
  TransceiveState *RxState;
  MsgType message;

  /* audio device driver */
  strcpy (dev_odo, ODO_DEVICE);		

  odo_frame = ODO_LENGTH * ODO_RATE * ODO_SIZE * ODO_CHANNELS / 8;
  buf = (unsigned char *) malloc(odo_frame);
  if (buf == NULL) {
    printf ("malloc fail: sizeof (buf) = %d\n", odo_frame);
    exit (1);
  }

  /* Initialization of receiver parameters */
  param = (Parameters *) init_RxVidParam (argc, argv);

  /* Initialization of audio */
  fd = BV_ODO_initialize (dev_odo);
  printf ("Initialized audio devices\n");

  /* Init. of network */
  RxState = (TransceiveState *) initialize_receive (param, AUDIO);

  /* loop for audio play until Control-C */
  for ( ; ; ) {		
    BV_ODO_receive (RxState, &message);
    memcpy (buf, message.data, odo_frame);

    wr = BV_ODO_play (fd, buf, odo_frame);
    printf ("Play size is %d\n", wr);

    /* wait for playback to complete before recording again */
    status = ioctl(fd, SOUND_PCM_SYNC, 0); 
    if (status == -1)
      perror("SOUND_PCM_SYNC ioctl failed");
  }
}



/* ***************************************************************
 * Function: void OLD_main (int argc, char **argv) 
 *
 *
 * *************************************************************** */
void OLD_main (int argc, char **argv)
{
  int   i;
  char aud_fname[80];

  Parameters *Tparam,
             *Rparam;

  pthread_t tid_VidTx,
            tid_VidRx,
            tid_AudRx, 
            tid_AudTx;

  pthread_attr_t attr;

  SEM_INIT (&VS, 0, 0);
  SEM_INIT (&VR, 0, 0);
  SEM_INIT (&AS, 0, 0);
  SEM_INIT (&AR, 0, 0);

  PTHREAD_ATTR_INIT (&attr);
  pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);


  for (i=1; i<argc; i++){
    switch (argv[i][1])
      {
      case 'i':
	//	strcpy(aud_fname, argv[++i]);
	//	if (audio_isaudiofile(aud_fname)) {
	//	  input_fd = open(aud_fname, O_RDONLY, 0666);
	//	  if (input_fd < 0)
	//	    fprintf(stderr,"could not open %s\n", aud_fname);
	//	} 
	//	else {
	//	  fprintf(stderr,"file %s is not an audio file\n", aud_fname);
	//	}
	break;
      case 'o':
	//	strcpy(aud_fname, argv[++i]);
	//	output_fd = open(aud_fname, O_WRONLY | O_CREAT | O_TRUNC, 0666);
	//	if (output_fd < 0)
	//	  fprintf(stderr,"could not create %s\n", aud_fname);
	break;
      case '?':
      default:
        usage(argv[0]);
        exit(0);
      }
  }

  /* Initialization of video devices */
  Tparam = (Parameters *) init_TxVidParam(argc, argv);
  Rparam = (Parameters *) init_RxVidParam(argc, argv);
  
  Tparam->data_via_file = Rparam->data_via_file = 0;
  Rparam->argc = argc;
  Rparam->argv = argv;

  //  sem_post (&VR);

  /* Creation of send/receive video threads */
  sleep (1);
  PTHREAD_CREATE (&tid_VidTx, &attr, Thread_TxVideo, (void *) Tparam);

  sleep (1);
  PTHREAD_CREATE (&tid_VidRx, &attr, Thread_RxVideo, (void *) Rparam);
printf ("main:1:OK\n");
  /* Initialization of audio devices */
  //  if (BV_init_audio ()) {
  //    fprintf (stderr, "Error: BV_init_audio\n");
  //    exit (0);
  //  }

  sleep (1);
  PTHREAD_CREATE (&tid_AudTx, &attr, Thread_TxAudio, (void *) Tparam);      

  sleep (1);
  PTHREAD_CREATE (&tid_AudRx, &attr, Thread_RxAudio, (void *) Rparam);

  pthread_exit ((void *) NULL);
}



/* *************************************************************************
 * Function    : void main (int argc, char **argv) 
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void main (int argc, char **argv)
{
  int   i;

  Parameters *Tparam,
             *Rparam;

  pthread_t tid_VidTx,
            tid_VidRx,
            tid_AudRx, 
            tid_AudTx;

  pthread_attr_t attr;

  SEM_INIT (&VS, 0, 0);
  SEM_INIT (&VR, 0, 0);
  SEM_INIT (&AS, 0, 0);
  SEM_INIT (&AR, 0, 0);

  PTHREAD_ATTR_INIT (&attr);
  pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);


  /* Process command line arguments */
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-C")) {
      if (!strcmp(argv[++i], "H261")) {
      	param->codec = H261;
      	param->cis_type = "H261";
      }
      else if (!strcmp(argv[i], "H263")) {
      	param->codec = H263;
      	param->cis_type = "H263";
      }
      else if (!strcmp(argv[i], "JPEG")) {
      	param->codec = JPEG;
      	param->cis_type = "Jpeg";
      }
      else if (!strcmp(argv[i], "MPEG")) {
      	param->codec = MPEG;
      	param->cis_type = "Mpeg1";
      }
      else if (!strcmp(argv[i], "RAW")) {
	param->codec = RAW;
      	param->cis_type = "RAW";
      }
      else {
	printf ("Specifiy your codec mode\n");
        usage(argv[0]);
	exit (0);
      }
    }
    if(!strcmp(argv[i], "-D")) {
      strcpy (devname, argv[++i]);
    }
    if (!strcmp(argv[i], "-i")) {
      strcpy (fname, argv[++i]);
      readable = 1;
      fp = fopen(fname, "r");
      if (fp == NULL) printf ("Can't open image file: %s\n", fname);
      else  printf ("Image file name is %s\n", fname);
    }
    if (!strcmp(argv[i], "-o")) {
      strcpy (fname, argv[++i]);
      writable = 1;
      fp = fopen(fname, "w");
      if (fp == NULL) printf ("Can't open image file: %s\n", fname);
      else  printf ("Image file name is %s\n", fname);
    }
    if(!strcmp(argv[i], "-r")) {
      //      rate = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-s")) {
      //      width = atoi(argv[++i]);
      //      height = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-v")) {
      verbose = 1;
    }
    if(!strcmp(argv[i], "-x")) {
      xwa.xdepth = atoi(argv[++i]);
      if (xwa.xdepth == 15)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB555;
      else if (xwa.xdepth == 16)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565;
      else if (xwa.xdepth == 24)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24;
      else if (xwa.xdepth == 32)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR32;
      else {
	printf("Unrecognized display depth. Type correct depth\n");
        usage(argv[0]);
	exit (0);
      }
    }
  }


  /* Initialization of video devices */
  Tparam = (Parameters *) init_TxVidParam(argc, argv);
  Rparam = (Parameters *) init_RxVidParam(argc, argv);
  
  Tparam->data_via_file = Rparam->data_via_file = 0;
  Rparam->argc = argc;
  Rparam->argv = argv;

  //  sem_post (&VR);

  /* Creation of send/receive video threads */
  sleep (1);
  PTHREAD_CREATE (&tid_VidTx, &attr, Thread_TxVideo, (void *) Tparam);

  sleep (1);
  PTHREAD_CREATE (&tid_VidRx, &attr, Thread_RxVideo, (void *) Rparam);
printf ("main:1:OK\n");
  /* Initialization of audio devices */
  //  if (BV_init_audio ()) {
  //    fprintf (stderr, "Error: BV_init_audio\n");
  //    exit (0);
  //  }

  sleep (1);
  PTHREAD_CREATE (&tid_AudTx, &attr, Thread_TxAudio, (void *) Tparam);      

  sleep (1);
  PTHREAD_CREATE (&tid_AudRx, &attr, Thread_RxAudio, (void *) Rparam);

  pthread_exit ((void *) NULL);
}

