/*	X Window Video Capture Testing Tool
 *
 *	gcc -o vdo -L/usr/X11R6/lib/ -lXt -lXaw -Wall vdo.c
 *
 */

/*  This program does not properly detect when the X Windows desktop is
 *  32 bits per pixel. (Some X functions use a value of 24 for the depth
 *  when it is really 32 for reasons I don't understand.) As a workaround,
 *  if you are running a 32-bpp X Windows screen, use -b on the command
 *  line, or uncomment this line:  */

/* video related  */
#include "vdo.h"

VIMAGE		vimage;
int		vid;
int		x_depth;
XtAppContext	app;
Widget		w_toplevel;
Widget		w_topbox;
Widget		w_menubox;
Widget		w_quit;
Widget		w_viewport;
Widget		w_stdbutton;
Widget		w_stdmenu;
Widget		w_sizebutton;
Widget		w_sizemenu;
struct v4l2_format	fmt;

struct { int width, height; } capturesize[] = {
  {160, 120}, {176, 144}, {192, 144}, {320, 240}, {384, 288}, {640, 480}, {768, 576}, {1024, 768},
};

/* *************************************************************************  
 * Function    :
 * Inputs      :
 * Outputs     :
 * Description :
 * ************************************************************************* */
__u32 frame_buffer_format(Display *d, Window w)
{
  __u32	fbf = V4L2_PIX_FMT_RGB565;

  XWindowAttributes wa;
  XGetWindowAttributes(d, w, &wa);

  if (wa.depth == 15)
    fbf = V4L2_PIX_FMT_RGB555;
  else if (wa.depth == 16)
    fbf = V4L2_PIX_FMT_RGB565;
  else if (wa.depth == 24)
    fbf = V4L2_PIX_FMT_BGR24;
  else if (wa.depth == 32)
    fbf = V4L2_PIX_FMT_BGR32;
  else {
    printf("Unrecognized display depth. You may get an X Windows error.\n");
  }

  return fbf;
}


/* *************************************************************************  
 * Function    :
 * Inputs      :
 * Outputs     :
 * Description :
 * 
 *
 * ************************************************************************* */
void stop_capturing()
{
  int	i;

  if (vid >= 0)	{
    i = V4L2_BUF_TYPE_CAPTURE;
    ioctl(vid, VIDIOC_STREAMOFF, &i);
    if (vimage.data)
      munmap(vimage.data, vimage.vidbuf.length);
    vimage.data = NULL;
    if (vimage.ximage)
      vimage.ximage = NULL;
  }
  
}


/* *************************************************************************  
 * Function    :
 * Inputs      :
 * Outputs     :
 * Description :
 * 
 *
 * ************************************************************************* */
int start_capturing()
{
  struct v4l2_requestbuffers req;
  int	err;
  Arg	arg[2];

#ifdef VDO_PIXELFORMAT
  fmt.fmt.pix.pixelformat = VDO_PIXELFORMAT;
#else
  fmt.fmt.pix.pixelformat = frame_buffer_format(XtDisplay(w_viewport),
				      DefaultRootWindow(XtDisplay(w_viewport)));
#endif
  err = ioctl(vid, VIDIOC_S_FMT, &fmt);
  if (err) {
    printf("S_FMT returned error %d\n",errno);
    return 1;
  }

  req.type = V4L2_BUF_TYPE_CAPTURE;
  err = ioctl(vid, VIDIOC_REQBUFS, &req);
  if (err < 0 || req.count < 1)	{
    printf("REQBUFS returned error %d, count %d\n", errno,req.count);
    return 1;
  }
  
  vimage.vidbuf.type = V4L2_BUF_TYPE_CAPTURE;
  err = ioctl(vid, VIDIOC_QUERYBUF, &(vimage.vidbuf));
  if (err < 0) {
    printf("QUERYBUF returned error %d\n",errno);
    return 1;
  }
  vimage.data = mmap(0, vimage.vidbuf.length, PROT_READ, MAP_SHARED, vid, vimage.vidbuf.offset);
  if ((int)vimage.data == -1) {
    printf("mmap() returned error %d\n", errno);
    return 1;
  }
  
  if ((err = ioctl(vid, VIDIOC_QBUF, &(vimage.vidbuf)))) {
    printf("QBUF returned error %d\n",errno);
    return 1;
  }
  err = ioctl(vid, VIDIOC_STREAMON, &(vimage.vidbuf.type));
  if (err)
    printf("STREAMON returned error %d\n",errno);

  vimage.ximage = XCreateImage(XtDisplay(w_viewport),
			       DefaultVisual(XtDisplay(w_viewport), 0),
			       (x_depth == 32) ? 24 : x_depth,
			       ZPixmap, 0, vimage.data,
			       fmt.fmt.pix.width, fmt.fmt.pix.height, 
			       16, 0);
  if (vimage.ximage == NULL) {
    printf("No XImage\n");
    return 1;
  }

  XtSetArg(arg[0], XtNwidth, max(160, fmt.fmt.pix.width));
  XtSetArg(arg[1], XtNheight, max(120, fmt.fmt.pix.height));
  XtSetValues(w_viewport, arg, 2);
  XtRealizeWidget(w_toplevel);

  return 0;
}


/* *************************************************************************  
 * Function    :
 * Inputs      :
 * Outputs     :
 * Description :
 * 
 *
 * ************************************************************************* */
void quit(Widget w, XtPointer client, XtPointer call)
{
  struct v4l2_performance	perf;

  if (vid >= 0)	{
    if (ioctl(vid, VIDIOC_G_PERF, &perf) != 0)
      printf("G_PERF returned error code %d\n",errno);
    else
      printf("Captured %d frames\n", perf.frames);
    stop_capturing();
    close(vid);
  }
  exit(0);
}


/* *************************************************************************  
 * Function    :
 * Inputs      :
 * Outputs     :
 * Description :
 * 
 *
 * ************************************************************************* */
void std_menu_callback(Widget w, XtPointer client, XtPointer call)
{
  StdMenuItem	*smi = (StdMenuItem *)client;

  stop_capturing();
  ioctl(vid, VIDIOC_S_STD, &smi->std);
  start_capturing();
}


/* *************************************************************************  
 * Function    :
 * Inputs      :
 * Outputs     :
 * Description :
 * 
 *
 * ************************************************************************* */
void size_menu_callback(Widget w, XtPointer client, XtPointer call)
{
  int	i = (int)client;

  fmt.fmt.pix.width = capturesize[i].width;
  fmt.fmt.pix.height = capturesize[i].height;
  stop_capturing();
  ioctl(vid, VIDIOC_S_FMT, &fmt);
  start_capturing();
}


/* *************************************************************************  
 * Function    : printctrl(struct v4l2_queryctrl *qc)
 * Inputs      : struct v4l2_queryctrl *qc
 * Outputs     : void
 * Description :
 * 
 *
 * ************************************************************************* */
void printctrl(struct v4l2_queryctrl *qc)
{	
  printf("  %-12s %-6s %8d %8d %7d %8d  %s\n", qc->name,
	 (qc->type == V4L2_CTRL_TYPE_MENU) ? "(menu)" : 
	 (qc->type == V4L2_CTRL_TYPE_BOOLEAN) ? "(bool)" : "(int)",
	 qc->minimum, qc->maximum, qc->step, qc->default_value, qc->group);
}

/* *************************************************************************
 * Function    : main (int argc, char *argv[])
 * Inputs      : int argc, char *argv[]
 * Outputs     : int 
 * Description :
 * 
 *
 * ************************************************************************* */
int main(int argc, char *argv[])
{
  char	my_device[64];
  int	err;
  int	i;
  int input;

  struct v4l2_capability cap;
  struct v4l2_streamparm parm;
  struct v4l2_queryctrl	qc;

  fd_set rdset;
  struct timeval timeout;
  int n;

  GC imagegc;
  int argi;
  struct v4l2_buffer	tempbuf;

/*->	Put in the device node name */
  strcpy(my_device, VDO_DEVICE);
  for (argi = 1; argi < argc; ++argi) {
    if (strcmp(argv[argi], "-b") == 0) {
      x_depth = 32;
      continue;
    }
    strcpy(my_device, argv[argi]);
  }
  x_depth = 16;

  vid = open(my_device, O_RDONLY);
  if (vid < 0) {
    printf("No video device \"%s\"\n", my_device);
    return 1;
  }

  err = ioctl(vid, VIDIOC_QUERYCAP, &cap);
  if (err) {
    printf("QUERYCAP returned error %d\n", errno);
    return 1;
  }
  printf("Device node:  %s\n", my_device);
  printf("Device:       %s\n", cap.name);
  printf("Capabilities:");
  if (cap.flags & V4L2_FLAG_READ)       printf(" capture via read();");
  if (cap.flags & V4L2_FLAG_STREAMING)  printf(" streaming capture;");
  if (cap.flags & V4L2_FLAG_PREVIEW)    printf(" automatic preview;");
  if (cap.flags & V4L2_FLAG_SELECT)     printf(" select() call;");
  if (cap.flags & V4L2_FLAG_TUNER)      printf(" tuner;");
  if (cap.flags & V4L2_FLAG_MONOCHROME) printf(" monochrome only;");
  printf("\n");
  if (cap.type != V4L2_TYPE_CAPTURE) {
    printf("Not a capture device.\n");
    return 1;
  }

  printf("Video inputs on this device:\n");
  for (i = 0, err = 0; err == 0; ++i) {	
    struct v4l2_input inp;
    inp.index = i;
    err = ioctl(vid, VIDIOC_ENUMINPUT, &inp);
    if (!err)
      printf("  %d: \"%s\" is a%s, %s audio\n", i, inp.name,
	    (inp.type == V4L2_INPUT_TYPE_TUNER) ? "TV tuner" : "n analog input",
	     (inp.capability & V4L2_INPUT_CAP_AUDIO) ? "has" : "no");
  }

  input = VDO_INPUT;
  err = ioctl(vid, VIDIOC_S_INPUT, &input);
  if (err)
    printf("S_INPUT returned error %d\n",errno);

  printf("Video standards supported:");
  for (i = 0, err = 0; err == 0; ++i) {
    struct v4l2_enumstd	estd;
    estd.index = i;
    err = ioctl(vid, VIDIOC_ENUMSTD, &estd);
    if (!err) {
      printf("%c %s", i?',':' ', estd.std.name);
    }
  }
  printf("\n");

  printf("Video capture image formats supported:\n");
  for (i = 0, err = 0; err == 0; ++i) {
    struct v4l2_fmtdesc	fmtd;

    fmtd.index = i;
    err = ioctl(vid, VIDIOC_ENUM_PIXFMT, &fmtd);
    if (!err)
      printf("  %d: %s  (%s)\n", i, fmtd.description,
	     (fmtd.flags & V4L2_FMT_FLAG_COMPRESSED) ?
	     "compressed" : "uncompressed");
  }

  printf("Controls supported:\n");
  printf("  Label        type    Minimum  Maximum    Step  Default Category\n");
  qc.id = V4L2_CID_BRIGHTNESS;
  if (ioctl(vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl(&qc);
  qc.id = V4L2_CID_CONTRAST;
  if (ioctl(vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl(&qc);
  qc.id = V4L2_CID_SATURATION;
  if (ioctl(vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl(&qc);
  qc.id = V4L2_CID_HUE;
  if (ioctl(vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl(&qc);

  if (!(cap.flags & V4L2_FLAG_STREAMING)) {
    printf("Device does not support streaming capture.\n");
    return 1;
  }

  parm.type = V4L2_BUF_TYPE_CAPTURE;
  ioctl(vid, VIDIOC_G_PARM, &parm);

  /* set proper framerate */
  parm.parm.capture.capturemode |= V4L2_CAP_TIMEPERFRAME;
  parm.parm.capture.timeperframe = VDO_FRAMERATE;

  err = ioctl(vid, VIDIOC_S_PARM, &parm);
  if (err)
    printf("S_PARM returned error %d\n",errno);

  fmt.type = V4L2_BUF_TYPE_CAPTURE;
  ioctl(vid, VIDIOC_G_FMT, &fmt);
  fmt.fmt.pix.width = VDO_WIDTH;
  fmt.fmt.pix.height = VDO_HEIGHT;
  ioctl(vid, VIDIOC_S_FMT, &fmt);

  w_toplevel = XtAppInitialize (&app, "xvideum",
				NULL, 0, 
				&argc, argv, 
				NULL, NULL, 0);
  w_topbox = XtVaCreateManagedWidget ("topbox",
				     boxWidgetClass, w_toplevel,
				     XtNorientation, XtorientVertical,
				     XtNborderWidth, 0,
				     XtNhSpace, 0,
				     XtNvSpace, 0,
				     NULL);
  w_menubox = XtVaCreateManagedWidget ("menubox",
				      boxWidgetClass, w_topbox,
				      XtNorientation, XtorientHorizontal,
				      XtNborderWidth, 0,
				      XtNhSpace, 4,
				      XtNvSpace, 1,
				      NULL);

  w_stdbutton = XtVaCreateManagedWidget ("menubutton",
					menuButtonWidgetClass, w_menubox,
					XtNlabel, "Standard",
					XtNmenuName, "standard",
					XtNborderWidth, 0,
					NULL);
  w_stdmenu = XtVaCreatePopupShell ("standard", /* same as XtNmenuName & static */
				    simpleMenuWidgetClass, w_stdbutton,
				    NULL);
  for (i = 0; ; ++i) {
    struct v4l2_enumstd	estd;
    estd.index = i;
    err = ioctl(vid, VIDIOC_ENUMSTD, &estd);
    if (err)
      break;
    stditem[i].std = estd.std;
    stditem[i].w_item = XtVaCreateManagedWidget (estd.std.name, 
						smeBSBObjectClass, w_stdmenu,
						 XtNlabel, estd.std.name, NULL);
    XtAddCallback (stditem[i].w_item, XtNcallback, std_menu_callback, &stditem[i]);
  }

  w_sizebutton = XtVaCreateManagedWidget ("menubutton",
					  menuButtonWidgetClass, w_menubox,
					  XtNlabel, "Zoom",
					  XtNmenuName, "Zoom",
					  XtNborderWidth, 0, NULL);
  w_sizemenu = XtVaCreatePopupShell ("Zoom", /* same as XtNmenuName & static */
				     simpleMenuWidgetClass, w_sizebutton, NULL);
  for (i = 0; i < sizeof(capturesize)/sizeof(capturesize[0]); ++i) {
    char sizestr[32];
    Widget w;

    sprintf(sizestr, "%d x %d", capturesize[i].width, capturesize[i].height);
    w = XtVaCreateManagedWidget (sizestr, smeBSBObjectClass, w_sizemenu,
				 XtNlabel, sizestr, NULL);
    XtAddCallback (w, XtNcallback, size_menu_callback, (char *)i);
  }

  w_quit = XtVaCreateManagedWidget ("Quit!", commandWidgetClass, w_menubox,
				    XtNborderWidth, 0, NULL);
  XtAddCallback (w_quit, XtNcallback, quit, NULL);
  w_viewport = XtVaCreateManagedWidget ("port",
					portholeWidgetClass, w_topbox,
					XtNwidth, VDO_WIDTH,
					XtNheight, VDO_HEIGHT,
					NULL);
  XtRealizeWidget(w_toplevel);
  imagegc = XCreateGC(XtDisplay(w_viewport), XtWindow(w_viewport), 0, NULL);
  if (imagegc == NULL) {
    printf("XCreateGC failed.\n");
    return 1;
  }

  if (start_capturing()) {
    return 1;
  }

  printf("Capturing %dx%dx%d \"%4.4s\" images\n", 
	 fmt.fmt.pix.width, fmt.fmt.pix.height, fmt.fmt.pix.depth, 
	 (char *)&fmt.fmt.pix.pixelformat);
  printf("Images are %d bytes each\n", fmt.fmt.pix.sizeimage);

  for (;;) {
    XEvent event;

    err = 0;
    if (XtIsRealized(w_viewport)) {
      FD_ZERO(&rdset);
      FD_SET(vid, &rdset);
      timeout.tv_sec = 1;
      timeout.tv_usec = 0;
      n = select(vid + 1, &rdset, NULL, NULL, &timeout);
      err = -1;
      if (n == -1)
	fprintf(stderr, "select error.\n");
      else if (n == 0)
	fprintf(stderr, "select timeout\n");
      else if (FD_ISSET(vid, &rdset))
	err = 0;

      if (err == 0) {
	tempbuf.type = vimage.vidbuf.type;
	err = ioctl(vid, VIDIOC_DQBUF, &tempbuf);
	if (err)
	  printf("DQBUF returned error %d\n", errno);

	XPutImage(XtDisplay(w_viewport), XtWindow(w_viewport), imagegc,
		  vimage.ximage, 0, 0, 0, 0,
		  fmt.fmt.pix.width, fmt.fmt.pix.height);

	err = ioctl(vid, VIDIOC_QBUF, &tempbuf);
	if (err)
	  printf("QBUF returned error %d\n", errno);
      }
    }
    else {
      XtAppPeekEvent(app, &event);
    }

    while (XtAppPending(app)) {
      XtAppNextEvent(app, &event);
      switch (event.type) {
      default:
	break;
      }
      XtDispatchEvent(&event);
    }
  }
  return 0;
}
