
/* *************************************************************************
 *  File : netmedia.c
 * -------------------------------------------------------------------------
 *  All rights are reserved by Bo-kyun Na, IntellActive Research Inc.
 * -------------------------------------------------------------------------
 * This example reads an image from /dev/video, transmits compressed video 
 * over the local network, and optionally displays the image on Xwindow.
 * ************************************************************************* */

/*  This program does not properly detect when the X Windows desktop is
 *  32 bits per pixel. (Some X functions use a value of 24 for the depth
 *  when it is really 32 for reasons I don't understand.) As a workaround,
 *  if you are running a 32-bpp X Windows screen, use -b on the command
 *  line, or uncomment this line:  */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

#include "params.h"		/* parameters */
#include "network.h"		/* network related */
#include "transceive.h"		/* parcket format related */

#include "vdo.h"		/* video related */
#include "h261.h"

unsigned char *clp;

unsigned long wpixel[3];
int expand;

extern int verbose;


/* *************************************************************************  
 * Function    :
 * Inputs      :
 * Outputs     :
 * Description :
 * ************************************************************************* */
__u32 frame_buffer_format(Display *d, Window w, XWA *xwa)
{
  __u32	fbf = V4L2_PIX_FMT_RGB565;

  XWindowAttributes wa;
  XGetWindowAttributes(d, w, &wa);

  xwa->xdepth = wa.depth;
  if (wa.depth == 15)
    fbf = V4L2_PIX_FMT_RGB555;
  else if (wa.depth == 16)
    fbf = V4L2_PIX_FMT_RGB565;
  else if (wa.depth == 24)
    fbf = V4L2_PIX_FMT_BGR24;
  else if (wa.depth == 32)
    fbf = V4L2_PIX_FMT_BGR32;
  else {
    printf("Unrecognized display depth. You may get an X Windows error.\n");
  }

  return fbf;
}


/* *************************************************************************  
 * Function    :
 * Inputs      :
 * Outputs     :
 * Description :
 * ************************************************************************* */
void stop_capturing(VIMAGE *vimage)
{
  if (vimage->vid >= 0)	{
    if (vimage->data) free(vimage->data);
    vimage->data = NULL;
    if (vimage->ximage) vimage->ximage = NULL;
  }
}


/* *************************************************************************  
 * Function    :
 * Inputs      :
 * Outputs     :
 * Description :
 * ************************************************************************* */
int start_capturing(VIMAGE *vimage, XWA *xwa, V4L2 *v4l2)
{
  int	err;
  Arg	arg[2];

  v4l2->fmt.fmt.pix.pixelformat = vimage->pixelfmt;

  err = ioctl(vimage->vid, VIDIOC_S_FMT, &v4l2->fmt);
  if (err) {
    printf("S_FMT returned error %d\n",errno);
    return 1;
  }

  vimage->data = malloc(v4l2->fmt.fmt.pix.sizeimage);
  if (vimage->data == NULL) {
    printf("malloc(%d) failed\n", v4l2->fmt.fmt.pix.sizeimage);
    return 1;
  }

  vimage->ximage = 
    XCreateImage (XtDisplay(xwa->viewport),
		  DefaultVisual(XtDisplay(xwa->viewport), 0),
		  (xwa->xdepth == 32) ? 24 : xwa->xdepth, ZPixmap, 0, 
		  vimage->data,
		  v4l2->fmt.fmt.pix.width, v4l2->fmt.fmt.pix.height, 16, 0);
  if (vimage->ximage == NULL) {
    printf("No XImage\n");
    return 1;
  }

  XtSetArg(arg[0], XtNwidth, max(160, v4l2->fmt.fmt.pix.width));
  XtSetArg(arg[1], XtNheight, max(120, v4l2->fmt.fmt.pix.height));
  XtSetValues(xwa->viewport, arg, 2);
  XtRealizeWidget(xwa->toplevel);

  return 0;
}


/* *************************************************************************  
 * Function    :
 * Inputs      :
 * Outputs     :
 * Description :
 * ************************************************************************* */
void quit_w(Widget w, XtPointer client, XtPointer call, VIMAGE *vimage)
{
  struct v4l2_performance	perf;

  if (vimage->vid >= 0)	{
    if (ioctl(vimage->vid, VIDIOC_G_PERF, &perf) != 0)
      printf("G_PERF returned error code %d\n",errno);
    else
      printf("Captured %d frames\n", perf.frames);
    stop_capturing(vimage);
    close(vimage->vid);
  }
  exit(0);
}


/* *************************************************************************  
 * Function    : std_menu_callback()
 * Inputs      : Widget w, XtPointer client, XtPointer call, 
 *  	         VIMAGE *vimage, XWA *xwa, V4L2 *v4l2
 * Outputs     :
 * Description :
 * ************************************************************************* */
void std_menu_callback(Widget w, XtPointer client, XtPointer call, 
		       VIMAGE *vimage, XWA *xwa, V4L2 *v4l2)
{
  StdMenuItem	*smi = (StdMenuItem *)client;

  stop_capturing(vimage);
  ioctl(vimage->vid, VIDIOC_S_STD, &smi->std);
  start_capturing(vimage, xwa, v4l2);
}


/* *************************************************************************  
 * Function    : size_menu_callback()
 * Inputs      : Widget w, XtPointer client, XtPointer call,
 *		 VIMAGE *vimage, XWA *xwa, V4L2 *v4l2
 * Outputs     :
 * Description :
 * ************************************************************************* */
void size_menu_callback(Widget w, XtPointer client, XtPointer call,
			VIMAGE *vimage, XWA *xwa, V4L2 *v4l2)
{
  int	i = (int)client;

  v4l2->fmt.fmt.pix.width = xwa->capturesize[i].width;
  v4l2->fmt.fmt.pix.height = xwa->capturesize[i].height;

  stop_capturing(vimage);
  ioctl(vimage->vid, VIDIOC_S_FMT, &v4l2->fmt);
  start_capturing(vimage, xwa, v4l2);
}


/* *************************************************************************  
 * Function    : printctrl(struct v4l2_queryctrl *qc)
 * Inputs      : struct v4l2_queryctrl *qc
 * Outputs     : void
 * Description :
 * ************************************************************************* */
void printctrl(struct v4l2_queryctrl *qc)
{	
  printf("  %-12s %-6s %8d %8d %7d %8d  %s\n", qc->name,
	 (qc->type == V4L2_CTRL_TYPE_MENU) ? "(menu)" : 
	 (qc->type == V4L2_CTRL_TYPE_BOOLEAN) ? "(bool)" : "(int)",
	 qc->minimum, qc->maximum, qc->step, qc->default_value, qc->group);
}

/* *************************************************************************
 * Function    : main (int argc, char *argv[])
 * Inputs      : int argc, char *argv[]
 * Outputs     : int 
 * Description : void *Thread_TxVideo (void *arg)
 * ************************************************************************* */
int main (int argc, char **argv)
{
  FILE *fp;
  char fname[80];
  int readable = 0, writable = 0;

  int i;
  int err;
  int input;
  struct v4l2_queryctrl qc;

  /* variables for network */
  Parameters *param;
  TransceiveState *tstate;

  /* variables for video */
  char	devname[64];
  //  unsigned char *capimg;	/* buffer for captured image */
  XWA xwa;
  V4L2 v4l2;
  VIMAGE vimage;

  StdMenuItem	stditem[MAXSTDS];

  Widget		w_topbox;
  Widget		w_menubox;
  Widget		w_quit;
  Widget		w_stdbutton;
  Widget		w_stdmenu;
  Widget		w_sizebutton;
  Widget		w_sizemenu;


  /* variables related to codec  */
  int first = 0;
  int CISsize;
  char inbuf[VDO_WIDTH * VDO_HEIGHT * 3/2]; 
  char outbuf[VDO_WIDTH * VDO_HEIGHT * 3/2];
  char tmpbuf[VDO_WIDTH * VDO_HEIGHT * 3/2];
  unsigned char *y, *u, *v;
  unsigned char *y1, *u1, *v1;
  int userno = 0;

  xwa.capturesize[0].width  = 176;
  xwa.capturesize[0].height = 144;

  xwa.capturesize[1].width  = 192;
  xwa.capturesize[1].height = 144;

  xwa.capturesize[2].width  = 320;
  xwa.capturesize[2].height = 240;

  xwa.capturesize[3].width  = 384;
  xwa.capturesize[3].height = 288;

  xwa.capturesize[4].width  = 640;
  xwa.capturesize[4].height = 480;

  xwa.capturesize[5].width  = 768;
  xwa.capturesize[5].height = 576;

  xwa.capturesize[6].width  = 1024;
  xwa.capturesize[6].height = 768;



  /* Put in the device node name */
  strcpy(devname, VDO_DEVICE);

  /* Initialization of video devices */
  param = (Parameters *) init_TxVidParam (argc, argv);

  /* Initialization of transmission parameters */
  if (param->transmit)
    tstate = (TransceiveState *) initialize_transmit(param, VIDEO);

  /* Pointers for the conversion of RGBtoYUV */
  y = inbuf;
  u = &inbuf[VDO_WIDTH * VDO_HEIGHT];
  v = &inbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  y1 = tmpbuf;
  u1 = &tmpbuf[VDO_WIDTH * VDO_HEIGHT];
  v1 = &tmpbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  xwa.xdepth = 16;		/* init. of depth on xwindow */      
  v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565;

  /* Process command line arguments */
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-C")) {
      if (!strcmp(argv[++i], "H261")) {
      	param->codec = H261;
      	param->cis_type = "H261";
      }
      else if (!strcmp(argv[i], "H263")) {
      	param->codec = H263;
      	param->cis_type = "H263";
      }
      else if (!strcmp(argv[i], "JPEG")) {
      	param->codec = JPEG;
      	param->cis_type = "Jpeg";
      }
      else if (!strcmp(argv[i], "MPEG")) {
      	param->codec = MPEG;
      	param->cis_type = "Mpeg1";
      }
      else if (!strcmp(argv[i], "RAW")) {
	param->codec = RAW;
      	param->cis_type = "RAW";
      }
      else {
	printf ("Specifiy your codec mode\n");
	exit (0);
      }
    }
    if(!strcmp(argv[i], "-D")) {
      strcpy (devname, argv[++i]);
    }
    if (!strcmp(argv[i], "-i")) {
      strcpy (fname, argv[++i]);
      readable = 1;
      fp = fopen(fname, "r");
      if (fp == NULL) printf ("Can't open image file: %s\n", fname);
      else  printf ("Image file name is %s\n", fname);
    }
    if (!strcmp(argv[i], "-o")) {
      strcpy (fname, argv[++i]);
      writable = 1;
      fp = fopen(fname, "w");
      if (fp == NULL) printf ("Can't open image file: %s\n", fname);
      else  printf ("Image file name is %s\n", fname);
    }
    if(!strcmp(argv[i], "-r")) {
      //      rate = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-s")) {
      //      width = atoi(argv[++i]);
      //      height = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-v")) {
      verbose = 1;
    }
    if(!strcmp(argv[i], "-x")) {
      xwa.xdepth = atoi(argv[++i]);
      if (xwa.xdepth == 15)
	vimage.pixelfmt = V4L2_PIX_FMT_RGB555;
      else if (xwa.xdepth == 16)
	vimage.pixelfmt = V4L2_PIX_FMT_RGB565;
      else if (xwa.xdepth == 24)
	vimage.pixelfmt = V4L2_PIX_FMT_BGR24;
      else if (xwa.xdepth == 32)
	vimage.pixelfmt = V4L2_PIX_FMT_BGR32;
      else {
	printf("Unrecognized display depth. Type correct depth\n");
	vimage.pixelfmt = 
	  frame_buffer_format(XtDisplay(xwa.viewport),
			      DefaultRootWindow(XtDisplay(xwa.viewport)), &xwa);
      }
    }
  }

  vimage.vid = open(devname, O_RDONLY);
  if (vimage.vid < 0) {
    printf("No video device \"%s\"\n", devname);
    return 1;
  }

  err = ioctl(vimage.vid, VIDIOC_QUERYCAP, &v4l2.cap);
  if (err) {
    printf("QUERYCAP returned error %d\n", errno);
    return 1;
  }
  printf("Device node:  %s\n", devname);
  printf("Device:       %s\n", v4l2.cap.name);
  printf("Capabilities:");
  if (v4l2.cap.flags & V4L2_FLAG_READ)       printf(" capture via read();");
  if (v4l2.cap.flags & V4L2_FLAG_STREAMING)  printf(" streaming capture;");
  if (v4l2.cap.flags & V4L2_FLAG_PREVIEW)    printf(" automatic preview;");
  if (v4l2.cap.flags & V4L2_FLAG_SELECT)     printf(" select() call;");
  if (v4l2.cap.flags & V4L2_FLAG_TUNER)      printf(" tuner;");
  if (v4l2.cap.flags & V4L2_FLAG_MONOCHROME) printf(" monochrome only;");
  printf("\n");
  if (v4l2.cap.type != V4L2_TYPE_CAPTURE) {
    printf("Not a capture device.\n");
    return 1;
  }

  printf("Video inputs on this device:\n");
  for (i = 0, err = 0; err == 0; ++i) {	
    struct v4l2_input inp;
    inp.index = i;
    err = ioctl(vimage.vid, VIDIOC_ENUMINPUT, &inp);
    if (!err)
      printf("  %d: \"%s\" is a%s, %s audio\n", i, inp.name,
	    (inp.type == V4L2_INPUT_TYPE_TUNER) ? "TV tuner" : "n analog input",
	     (inp.capability & V4L2_INPUT_CAP_AUDIO) ? "has" : "no");
  }

  input = VDO_INPUT;
  err = ioctl(vimage.vid, VIDIOC_S_INPUT, &input);
  if (err)
    printf("S_INPUT returned error %d\n",errno);

  printf("Video standards supported:");
  for (i = 0, err = 0; err == 0; ++i) {
    struct v4l2_enumstd	estd;
    estd.index = i;
    err = ioctl(vimage.vid, VIDIOC_ENUMSTD, &estd);
    if (!err) {
      printf("%c %s", i?',':' ', estd.std.name);
    }
  }
  printf("\n");

  printf("Video capture image formats supported:\n");
  for (i = 0, err = 0; err == 0; ++i) {
    struct v4l2_fmtdesc	fmtd;

    fmtd.index = i;
    err = ioctl(vimage.vid, VIDIOC_ENUM_PIXFMT, &fmtd);
    if (!err)
      printf("  %d: %s  (%s)\n", i, fmtd.description,
	     (fmtd.flags & V4L2_FMT_FLAG_COMPRESSED) ?
	     "compressed" : "uncompressed");
  }

  printf("Controls supported:\n");
  printf("  Label        type    Minimum  Maximum    Step  Default Category\n");
  qc.id = V4L2_CID_BRIGHTNESS;
  if (ioctl(vimage.vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl(&qc);
  qc.id = V4L2_CID_CONTRAST;
  if (ioctl(vimage.vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl(&qc);
  qc.id = V4L2_CID_SATURATION;
  if (ioctl(vimage.vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl(&qc);
  qc.id = V4L2_CID_HUE;
  if (ioctl(vimage.vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl(&qc);

  if (!(v4l2.cap.flags & V4L2_FLAG_READ)) {
    printf("Device does not support the read() call.\n");
    return 1;
  }

  v4l2.parm.type = V4L2_BUF_TYPE_CAPTURE;
  ioctl(vimage.vid, VIDIOC_G_PARM, &v4l2.parm);

  /* set proper framerate */
  v4l2.parm.parm.capture.capturemode |= V4L2_CAP_TIMEPERFRAME;
  v4l2.parm.parm.capture.timeperframe = VDO_FRAMERATE;

  err = ioctl(vimage.vid, VIDIOC_S_PARM, &v4l2.parm);
  if (err)
    printf("S_PARM returned error %d\n",errno);

  v4l2.fmt.type = V4L2_BUF_TYPE_CAPTURE;
  ioctl(vimage.vid, VIDIOC_G_FMT, &v4l2.fmt);
  v4l2.fmt.fmt.pix.width = VDO_WIDTH;
  v4l2.fmt.fmt.pix.height = VDO_HEIGHT;
  ioctl(vimage.vid, VIDIOC_S_FMT, &v4l2.fmt);

  xwa.toplevel = XtAppInitialize (&xwa.app, "xvideum",
				NULL, 0, 
				&argc, argv, 
				NULL, NULL, 0);
  w_topbox = XtVaCreateManagedWidget ("topbox",
				     boxWidgetClass, xwa.toplevel,
				     XtNorientation, XtorientVertical,
				     XtNborderWidth, 0,
				     XtNhSpace, 0,
				     XtNvSpace, 0,
				     NULL);
  w_menubox = XtVaCreateManagedWidget ("menubox",
				      boxWidgetClass, w_topbox,
				      XtNorientation, XtorientHorizontal,
				      XtNborderWidth, 0,
				      XtNhSpace, 4,
				      XtNvSpace, 1,
				      NULL);

  w_stdbutton = XtVaCreateManagedWidget ("menubutton",
					menuButtonWidgetClass, w_menubox,
					XtNlabel, "Standard",
					XtNmenuName, "standard",
					XtNborderWidth, 0,
					NULL);
  w_stdmenu = XtVaCreatePopupShell ("standard", /* same as XtNmenuName & static */
				    simpleMenuWidgetClass, w_stdbutton,
				    NULL);
  for (i = 0; ; ++i) {
    struct v4l2_enumstd	estd;
    estd.index = i;
    err = ioctl(vimage.vid, VIDIOC_ENUMSTD, &estd);
    if (err)
      break;
    stditem[i].std = estd.std;
    stditem[i].w_item = XtVaCreateManagedWidget (estd.std.name, 
						smeBSBObjectClass, w_stdmenu,
						 XtNlabel, estd.std.name, NULL);
    XtAddCallback (stditem[i].w_item, XtNcallback, 
		   (XtCallbackProc) std_menu_callback, &stditem[i]);
  }

  w_sizebutton = XtVaCreateManagedWidget ("menubutton",
					  menuButtonWidgetClass, w_menubox,
					  XtNlabel, "Zoom",
					  XtNmenuName, "Zoom",
					  XtNborderWidth, 0, NULL);
  w_sizemenu = XtVaCreatePopupShell ("Zoom", /* same as XtNmenuName & static */
				     simpleMenuWidgetClass, w_sizebutton, NULL);
  for (i = 0; i < sizeof(xwa.capturesize)/sizeof(xwa.capturesize[0]); ++i) {
    char sizestr[32];
    Widget w;

    //    sprintf(sizestr, "%d x %d", 
    //	    xwa.capturesize[i].width, xwa.capturesize[i].height);
    sprintf(sizestr, "%d", xwa.capturesize[i].width);
    w = XtVaCreateManagedWidget (sizestr, smeBSBObjectClass, w_sizemenu,
				 XtNlabel, sizestr, NULL);
    XtAddCallback (w, XtNcallback, (XtCallbackProc) size_menu_callback, (char *)i);
  }

  w_quit = XtVaCreateManagedWidget ("Quit!", commandWidgetClass, w_menubox,
				    XtNborderWidth, 0, NULL);
  XtAddCallback (w_quit, XtNcallback, (XtCallbackProc) quit_w, NULL);
  xwa.viewport = XtVaCreateManagedWidget ("port",
					portholeWidgetClass, w_topbox,
					XtNwidth, VDO_WIDTH,
					XtNheight, VDO_HEIGHT,
					NULL);
  XtRealizeWidget(xwa.toplevel);

  xwa.gc = XCreateGC(XtDisplay(xwa.viewport), XtWindow(xwa.viewport), 0, NULL);
  if (xwa.gc == NULL) {
    printf("XCreateGC failed.\n");
    return 1;
  }

  if (start_capturing(&vimage, &xwa, &v4l2)) {
    return 1;
  }

  printf("Capturing %dx%dx%d \"%4.4s\" images\n", 
	 v4l2.fmt.fmt.pix.width, v4l2.fmt.fmt.pix.height, 
	 v4l2.fmt.fmt.pix.depth, (char *)&v4l2.fmt.fmt.pix.pixelformat);
  printf("Images are %d bytes each\n", v4l2.fmt.fmt.pix.sizeimage);


  for (;;) {
    XEvent event;

    err = 0;
    if (XtIsRealized(xwa.viewport)) {
      read(vimage.vid, vimage.data, v4l2.fmt.fmt.pix.sizeimage);
      XPutImage(XtDisplay(xwa.viewport), XtWindow(xwa.viewport), xwa.gc,
		vimage.ximage, 0, 0, 0, 0, 
		v4l2.fmt.fmt.pix.width, v4l2.fmt.fmt.pix.height);

    }
    else {
      XtAppPeekEvent(xwa.app, &event);
    }

    while (XtAppPending(xwa.app)) {
      XtAppNextEvent(xwa.app, &event);
      switch (event.type) {
      default:
	break;
      }
      XtDispatchEvent(&event);
    }
  }
  return 0;
}
