
/* *************************************************************************
 *  File : netmedia.c
 * -------------------------------------------------------------------------
 *  All rights are reserved by Bo-kyun Na, IntellActive Research Inc.
 * -------------------------------------------------------------------------
 * This example reads an image from /dev/video, transmits compressed video 
 * over the local network, and optionally displays the image on Xwindow.
 * ************************************************************************* */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

#include "params.h"
#include "network.h"
#include "transceive.h"
#include "vdo.h"
#include "h261.h"

unsigned char *clp;
unsigned long wpixel[3];

extern int verbose;


/* *************************************************************************
 * Function    : main (int argc, char *argv[])
 * Inputs      : int argc, char *argv[]
 * Outputs     : int 
 * Description :
 * 
 *
 * ************************************************************************* */
int main(int argc, char *argv[])
{
  FILE *fp;
  char fname[80];
  int readable = 0, writable = 0;

  int i;
  int	err;

  Parameters *param;
  TransceiveState *tstate;

  int vid;
  char	devname[64];
  unsigned char *capimg;
  XWA xwa;
  V4L2 v4l2;
  VIMAGE vimage;


  /*->	Put in the device node name */
  strcpy(my_device, VDO_DEVICE);

  /* Process command line arguments */
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-C")) {
      if (!strcmp(argv[++i], "H261")) {
	param->codec = H261;
	param->cis_type = "H261";
      }
      else if (!strcmp(argv[i], "H263")) {
	param->codec = H263;
	param->cis_type = "H263";
      }
      else if (!strcmp(argv[i], "JPEG")) {
	param->codec = JPEG;
	param->cis_type = "Jpeg";
      }
      else if (!strcmp(argv[i], "MPEG")) {
	param->codec = MPEG;
	param->cis_type = "Mpeg1";
      }
      else if (!strcmp(argv[i], "RAW")) {
	param->codec = RAW;
	param->cis_type = "RAW";
      }
      else {
	printf ("Specifiy your codec mode\n");
	exit (0);
      }
    }
    if(!strcmp(argv[i], "-D")) {
      strcpy (devname, argv[++i]);
    }
    if (!strcmp(argv[i], "-i")) {
      strcpy (fname, argv[++i]);
      readable = 1;
      fp = fopen(fname, "r");
      if (fp == NULL) printf ("Can't open image file: %s\n", fname);
      else  printf ("Image file name is %s\n", fname);
    }
    if (!strcmp(argv[i], "-o")) {
      strcpy (fname, argv[++i]);
      writable = 1;
      fp = fopen(fname, "w");
      if (fp == NULL) printf ("Can't open image file: %s\n", fname);
      else  printf ("Image file name is %s\n", fname);
    }
    if(!strcmp(argv[i], "-r")) {
      v4l2.parm.parm.capture.timeperframe = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-s")) {
      v4l2.fmt.fmt.pix.width = atoi(argv[++i]);
      v4l2.fmt.fmt.pix.height = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-v")) {
      verbose = 1;
    }
    if(!strcmp(argv[i], "-x")) {
      xwa.xdepth = atoi(argv[++i]);
      if (xwa.xdepth == 15)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB555;
      else if (xwa.xdepth == 16)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565;
      else if (xwa.xdepth == 24)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24;
      else if (xwa.xdepth == 32)
	v4l2.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR32;
      else {
	printf("Unrecognized display depth. Type correct depth\n");
	exit (0);
      }
    }
  }

  BV_init_capture_vdoframe ();
  BV_init_display ();

  if (start_capturing()) {
    return 1;
  }

  for (;;) {
    BV_get_vdoframe ();
    BV_display_vdoframe ();
    if (ioctl(vid, VIDIOC_QBUF, &tempbuf))
      printf("QBUF returned error %d\n", errno);
  }

}


